import re

from collections import namedtuple
from functools import partial
from psycopg2._psycopg import ProgrammingError
from psycopg2.extensions import AsIs, adapt, new_type, new_array_type, register_type, STATUS_IN_TRANSACTION
from psycopg2.extras import _solve_conn_curs

from spans.types import range_
from spans import floatrange

__all__ = [
    "raw_range",
    "empty_rawrange",
    "floatrange_preprocess",
    "parse_raw_range_string",
    "cast_raw_range_string",
    "register_range_caster",
    "adapt_range",
    "query_range_oids"
]

_raw_range = namedtuple("rawrange", ["lower", "upper", "lower_inc", "upper_inc", "empty"])
raw_range = partial(_raw_range, empty=False)
empty_rawrange = _raw_range(*([None] * 4), empty=True)

try:
    ustr = unicode
except NameError:
    ustr = str

def _raw_token(scanner, token):
    return token

def _unquote_token(scanner, token):
    return token[1:-1].encode().decode("unicode_escape")

raw_range_scanner = re.Scanner([
    (r"(['\"])(?:[^\1\\]|\\.)*?\1", _unquote_token),
    (r"(\[|\]|\(|\)|,)", _raw_token),
    (r"[^\[\]\(\),\s]+", _raw_token),
    (r"\s+", None)
])

def parse_raw_range_string(range_):
    # An empty range needs special handling
    if range_ == "empty":
        return empty_rawrange

    # Try to parse string using regex scanner
    tokens = raw_range_scanner.scan(range_)[0]

    # NOTE: This assumes that the given range is well formed

    return raw_range(
        lower=tokens[1] if tokens[2] == "," else None,
        upper=tokens[-2] if tokens[-3] == "," else None,
        lower_inc=True if tokens[0] == "[" else False,
        upper_inc=True if tokens[-1] == "]" else False)

def cast_raw_range_string(pyrange, range_, cur=None, subtype_oid=None):
    # If we got NULL just roll with it and return None
    if range_ is None:
        return None

    raw_range = parse_raw_range_string(range_)

    if raw_range.empty:
        return pyrange.empty()

    lower = raw_range.lower
    upper = raw_range.upper

    # Try to cast the individual components if cursor and subtype oid are
    # provided. This should be the standard case
    if cur is not None and subtype_oid is not None:
        lower = cur.cast(subtype_oid, lower)
        upper = cur.cast(subtype_oid, upper)

    return pyrange(lower, upper, raw_range.lower_inc, raw_range.upper_inc)

def register_range_caster(pgrange, pyrange, oid, subtype_oid, array_oid, scope=None):
    # Create an adapter for this range type
    adapter = partial(cast_raw_range_string, pyrange, subtype_oid=subtype_oid)

    # Create types using adapter
    range_type = new_type((oid,), pgrange, adapter)
    range_array_type = new_array_type((array_oid,), pgrange, range_type)

    register_type(range_type, scope)
    register_type(range_array_type, scope)

def adapt_range(pgrange, pyrange):
    if isinstance(pgrange, ustr):
        pgrange = pgrange.encode("utf8")

    if not isinstance(pyrange, range_):
        raise ValueError((
            "Trying to adapt range {range.__class__.__name__} which does not "
            "extend base range type.").format(range=pyrange))

    if not pyrange:
        return AsIs((b"'empty'::" + pgrange).decode("utf8"))

    lower = b"NULL"
    if not pyrange.lower_inf:
        lower = adapt(pyrange.lower).getquoted()

    upper = b"NULL"
    if not pyrange.upper_inf:
        upper = adapt(pyrange.upper).getquoted()

    return AsIs(b"".join([
        pgrange,
        b"(",
        lower,
        b", ",
        upper,
        b", '",
        b"[" if pyrange.lower_inc else b"(",
        b"]" if pyrange.upper_inc else b")",
        b"')",
    ]).decode("utf8"))

def query_range_oids(pgrange, conn_or_curs):
    conn, curs = _solve_conn_curs(conn_or_curs)

    # Store transaction status
    conn_status = conn.status

    # Calculate absolutre location for pgrange
    if "." in pgrange:
        schema, name = pgrange.split(".", 1)
    else:
        name = pgrange
        schema = "public"

    # Query oids
    try:
        curs.execute("""
            SELECT
                r.rngtypid,
                r.rngsubtype,
                t.typarray
            FROM pg_range r
                JOIN pg_type t ON(t.oid = r.rngtypid)
                JOIN pg_namespace ns ON(ns.oid = t.typnamespace)
            WHERE t.typname = %s AND ns.nspname = %s;
        """, (name, schema))
    except ProgrammingError:
        if not conn.autocommit:
            conn.rollback()
        raise

    type_data = curs.fetchone()

    # Reset status if it was tampered with
    if conn_status != STATUS_IN_TRANSACTION and not conn.autocommit:
        conn.rollback()

    if type_data is None:
        raise ProgrammingError("PostgreSQL type '{type}' not found".format(
            type=name))

    return type_data

def floatrange_preprocess(lower=None, upper=None, lower_inc=None, upper_inc=None):
    if lower is not None:
        lower = float(lower)

    if upper is not None:
        upper = float(upper)

    return floatrange(lower, upper, lower_inc, upper_inc)



import datetime as dt
import os

from functools import partial
from psycopg2._psycopg import ProgrammingError, InternalError
from spans import *
from unittest import TestCase, TestSuite, defaultTestLoader, skipUnless

from . import *
from ._utils import *

def build_test_dsn():
    dbname = os.environ.get("PSYCOSPANS_TESTDB", "psycospans_test")
    dbhost = os.environ.get("PSYCOSPANS_TESTDB_HOST", None)
    dbport = os.environ.get("PSYCOSPANS_TESTDB_PORT", None)
    dbuser = os.environ.get("PSYCOSPANS_TESTDB_USER", None)
    dbpass = os.environ.get("PSYCOSPANS_TESTDB_PASSWORD", None)

    dsn = ["dbname=" + dbname]

    if dbhost is not None:
        dsn.append("host=" + dbhost)

    if dbport is not None:
        dsn.append("port=" + dbport)

    if dbuser is not None:
        dsn.append("user=" + dbuser)

    if dbpass is not None:
        dsn.append("password=" + dbpass)

    return " ".join(dsn)

def can_connect():
    try:
        cnn = connect(build_test_dsn())
    except:
        return False
    else:
        cnn.close()
        return True

class TestCasting(TestCase):
    def test_parse_raw_range_string(self):
        self.assertEqual(
            parse_raw_range_string("empty"), empty_rawrange)

        self.assertEqual(
            parse_raw_range_string("[,)"), raw_range(None, None, True, False))
        self.assertEqual(
            parse_raw_range_string("(,)"), raw_range(None, None, False, False))
        self.assertEqual(
            parse_raw_range_string("(,]"), raw_range(None, None, False, True))
        self.assertEqual(
            parse_raw_range_string("[,]"), raw_range(None, None, True, True))

        self.assertEqual(
            parse_raw_range_string("[1, )"), raw_range("1", None, True, False))
        self.assertEqual(
            parse_raw_range_string("[, 2]"), raw_range(None, "2", True, True))
        self.assertEqual(
            parse_raw_range_string("(3, 4)"), raw_range("3", "4", False, False))
        self.assertEqual(
            parse_raw_range_string("[3.5, 4.7)"), raw_range("3.5", "4.7", True, False))

        self.assertEqual(
            parse_raw_range_string(r"['\\, \' \\', 'test']"),
            raw_range("\\, ' \\", "test", True, True))
        self.assertEqual(
            parse_raw_range_string(r"['\\, \"\' \\', 1]"),
            raw_range("\\, \"' \\", "1", True, True))
        self.assertEqual(
            parse_raw_range_string(r"[1, '\\, \' \\']"),
            raw_range("1", "\\, ' \\", True, True))

    def test_adapt_range(self):
        with self.assertRaises(ValueError):
            adapt_range("int4range", None)

        self.assertEqual(
            adapt_range("int4range", intrange(1, 5)).getquoted(),
            b"int4range(1, 5, '[)')")

        self.assertEqual(
            adapt_range("numrange", floatrange(2.5, 10.25, False)).getquoted(),
            b"numrange(2.5, 10.25, '()')")

        self.assertEqual(
            adapt_range("daterange", daterange(
                dt.date(2013, 1, 1), dt.date(2013, 2, 15))).getquoted(),
            b"daterange('2013-01-01'::date, '2013-02-15'::date, '[)')")

        self.assertEqual(
            adapt_range(
                "tsrange",
                datetimerange(upper=dt.datetime(2013, 1, 1, 4, 50))).getquoted(),
            b"tsrange(NULL, '2013-01-01T04:50:00'::timestamp, '()')")

        self.assertEqual(
            adapt_range("int4range", intrange.empty()).getquoted(),
            b"'empty'::int4range")

@skipUnless(can_connect(), "No test database available")
class TestDatabase(TestCase):
    def setUp(self):
        self.conn = connect(build_test_dsn())

    def tearDown(self):
        self.conn.close()

    def test_intrange(self):
        cur = self.conn.cursor()
        test_range = intrange(1, 10)
        cur.execute("SELECT int4range(5, NULL), %s", (test_range,))

        self.assertEqual(
            cur.fetchone(),
            (intrange(5), test_range))

    def test_floatrange(self):
        cur = self.conn.cursor()
        test_range = floatrange(0.5, lower_inc=False)
        cur.execute("SELECT numrange(5.5, 10.25), %s", (test_range,))

        self.assertEqual(
            cur.fetchone(),
            (floatrange(5.5, 10.25), test_range))

    def test_daterange(self):
        cur = self.conn.cursor()
        test_range = daterange(dt.date(2013, 1, 1), dt.date(2013, 2, 15))
        cur.execute("SELECT daterange('2013-04-04'::date, NULL), %s", (test_range,))

        self.assertEqual(
            cur.fetchone(),
            (daterange(dt.date(2013, 4, 4)), test_range))

    def test_datetimerange(self):
        cur = self.conn.cursor()
        test_range = datetimerange(
            dt.datetime(2013, 1, 1, 10, 20, 30),
            dt.datetime(2013, 2, 15, 20, 30, 40),
            upper_inc=True)
        cur.execute("SELECT tsrange('2013-04-04 04:50:45'::timestamp, NULL), %s", (test_range,))

        self.assertEqual(
            cur.fetchone(),
            (datetimerange(dt.datetime(2013, 4, 4, 4, 50, 45)), test_range))

    def test_query_range_oids(self):
        q = partial(query_range_oids, conn_or_curs=self.conn)

        self.assertEqual(q("pg_catalog.int4range"), (3904, 23, 3905))
        self.assertEqual(q("pg_catalog.int8range"), (3926, 20, 3927))
        self.assertEqual(q("pg_catalog.numrange"), (3906, 1700, 3907))
        self.assertEqual(q("pg_catalog.daterange"), (3912, 1082, 3913))
        self.assertEqual(q("pg_catalog.tsrange"), (3908, 1114, 3909))

    def test_custom_range(self):
        cur = self.conn.cursor()

        # This may fail if the type already exists
        try:
            cur.execute("CREATE TYPE intervalrange AS RANGE(SUBTYPE = interval)")
        except ProgrammingError:
            self.conn.rollback()
        else:
            self.conn.commit()

        register_range_type("intervalrange", timedeltarange, self.conn)

        test_range = timedeltarange(dt.timedelta(1, 5), dt.timedelta(1, 10))
        cur.execute("SELECT intervalrange('1 day'::interval, NULL), %s", (test_range,))

        self.assertEqual(
            cur.fetchone(),
            (timedeltarange(dt.timedelta(1)), test_range))

        # This may fail if the type already existed before test
        try:
            cur.execute("DROP TYPE intervalrange")
        except InternalError:
            self.conn.rollback()
        else:
            self.conn.commit()
        self.conn.commit()

def suite():
    return TestSuite(map(defaultTestLoader.loadTestsFromTestCase, [
        TestCasting,
        TestDatabase
    ]))



from functools import partial, wraps

from psycopg2 import connect as _connect
from psycopg2.extensions import register_adapter
from spans import *

from ._utils import adapt_range, register_range_caster, query_range_oids, floatrange_preprocess

__version__ = "0.1.1"

__all__ = [
    "connect",
    "register_range_type"
]

@wraps(_connect)
def connect(*keys, **kwargs):
    conn = _connect(*keys, **kwargs)

    if conn.server_version < 90200:
        raise RuntimeError("Range types not available in version {version}".format(
            conn.server_version))

    # Register range types
    register_range_caster(
        "int4range", intrange, oid=3904, subtype_oid=23, array_oid=3905, scope=conn)
    register_range_caster(
        "int8range", intrange, oid=3926, subtype_oid=20, array_oid=3927, scope=conn)
    register_range_caster(
        "numrange", floatrange_preprocess, oid=3906, subtype_oid=1700, array_oid=3907, scope=conn)
    register_range_caster(
        "daterange", daterange, oid=3912, subtype_oid=1082, array_oid=3913, scope=conn)
    register_range_caster(
        "tsrange", datetimerange, oid=3908, subtype_oid=1114, array_oid=3909, scope=conn)

    return conn

def register_range_type(pgrange, pyrange, conn):
    """
    Register a new range type as a PostgreSQL range.

        >>> register_range_type("int4range", intrange, conn)

    The above will make sure intrange is regarded as an int4range for queries
    and that int4ranges will be cast into intrange when fetching rows.

    pgrange should be the full name including schema for the custom range type.

    Note that adaption is global, meaning if a range type is passed to a regular
    psycopg2 connection it will adapt it to its proper range type. Parsing of
    rows from the database however is not global and just set on a per connection
    basis.
    """

    register_adapter(pyrange, partial(adapt_range, pgrange))
    register_range_caster(
        pgrange, pyrange, *query_range_oids(pgrange, conn), scope=conn)

register_adapter(intrange, partial(adapt_range, "int4range"))
register_adapter(floatrange, partial(adapt_range, "numrange"))
register_adapter(daterange, partial(adapt_range, "daterange"))
register_adapter(datetimerange, partial(adapt_range, "tsrange"))
